/*	$LAAS: $ */

/* 
 * Copyright (c) 2000-2004 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#ifdef VXWORKS
#include <vxWorks.h>
#endif

#if defined(__RTAI__) && defined(__KERNEL__)
# include <linux/kernel.h>
# include <linux/sched.h>
#else
# include <stdio.h>
# include <string.h>
#endif

#include <math.h>

#include "libt3d.h"
#include "t3d.h"


/* Local functions --------------------------------------------------- */

static void	compMatrices(T3D *a, T3D *b);


/*
 * t3dComp ------------------------------------------------------------
 *
 * Compose T3D a with T3D b, store the result in a and return a.
 * If a represents the transformation from a frame A to a frame X and b
 * represents the transformation from a frame X to a frame B, the result of
 * this fonction is a transformation from the frame A to the frame B.
 * Or for short : t3Comp("A_2_X", "X_2_B") = "A_2_B".
 */

T3D *
t3dComp(T3D *a, T3D *b)
{
   if (!t3dOk(a) || !t3dOk(b)) { t3dWarn(); return NULL; }
   
   /* Handle all type compositions through the T3D_MATRIX
    * type since the direct function would require to compute the
    * matrices anyways. */ 

   if (a->type != T3D_MATRIX) {
      if (a->any.flags & T3D_ALLOW_CONVERSION)
	 t3dConvertTo(T3D_MATRIX, a);
      else {
	 a->any.err = T3D_ENOTIMPLEMENTED;
	 t3dWarn();
	 return NULL;
      }
   }

   if (b->type != T3D_MATRIX) {
      if (b->any.flags & T3D_ALLOW_CONVERSION)
	 t3dConvertTo(T3D_MATRIX, b);
      else {
	 b->any.err = T3D_ENOTIMPLEMENTED;
	 t3dWarn();
	 return NULL;
      }
   }

   /* now compute the result. a and b are T3D_MATRIX */
   compMatrices(a, b);
   return t3dOk(a)? a : NULL;
}


/*
 * t3dCompIn ------------------------------------------------------------
 *
 * Compose T3D a with T3D b, store the result in c and return c.
 * See t3dComp().
 */

T3D *
t3dCompIn(T3D *c, T3D *a, T3D *b)
{
   return t3dComp(t3dCopyIn(c, a), b);
}


/*
 * --- compMatrices -----------------------------------------------------
 *
 * Compose two T3D_MATRIX, and variances on Euler parameters.
 */

static void
compMatrices(T3D *a, T3D *b)
{ 
   double c[16];			/* resulting matrix */
   double eulera[3], eulerb[3];		/* Euler angles of a and b */
   double rota[9], rotb[9];		/* rotation matrices of a and b */

   double d, d2;
   double Jee1[36], Jee2[36];
   double JeaR[27], JRea[27];
   double Jeaea1[9], Jeaea2[9];
   double JRa[9], JRb[9], JRg[9];
   double Jta[3], Jtb[3], Jtg[3];
   double Jtea2[9];
   double tmp1[21], tmp2[21];

   /* compute Euler angles of a and b */
   t3dEulerAnglesOfMatrix(a->matrix.matrix, eulera);
   t3dEulerAnglesOfMatrix(b->matrix.matrix, eulerb);

   /* extract rotation matrices from a and b */
   t3dRotationOfMatrix(a->matrix.matrix, rota);
   t3dRotationOfMatrix(b->matrix.matrix, rotb);

   /* compose matrices */
   t3d_dm_mult(b->matrix.matrix, a->matrix.matrix, c, 4, 4, 4, 4);


   /*
    * Now variances. If ea are the Euler angles, we just say that 
    *     Dea = JeaR JRea1 Dea1 + JeaR JRea2 Dea2
    */


   /* first, the jacobian JeaR */
   d = sqrt(d2 = (c[0] * c[0] + c[4] * c[4]));

   if (d2 > 1e-5) {
      double s = c[10] * c[10] + c[9] * c[9];
      t3d_dm_fill(
	 JeaR, 3, 9,
	 -c[4]/d2,    0.0, 0.0, c[0]/d2,     0.0, 0.0, 0.0, 0.0,     0.0,
	 c[8]*c[0]/d, 0.0, 0.0, c[8]*c[4]/d, 0.0, 0.0, -d,  0.0,     0.0,
	 0.0,         0.0, 0.0, 0.0,         0.0, 0.0, 0.0, c[10]/s, -c[9]/s
	 );

   } else
      t3d_dm_fill(
	 JeaR, 3, 9,
	 0.0,         -c[5], 0.0, 0.0,         c[1], 0.0, 0.0, 0.0, 0.0,
	 c[8]*c[0]/d, 0.0,   0.0, c[8]*c[4]/d, 0.0,  0.0, -d,  0.0, 0.0,
	 0.0,         0.0,   0.0, 0.0,         0.0,  0.0, 0.0, 0.0, 0.0
	 );
  
   /* then, the two from the rotation matrices  R = R2 R1 */

   t3dDerivativeOfEuler(eulera, JRa, JRb, JRg);
   t3d_dm_mult(rotb, JRa, tmp1, 3, 3, 3, 3); memcpy(JRa, tmp1, sizeof(JRa));
   t3d_dm_mult(rotb, JRb, tmp1, 3, 3, 3, 3); memcpy(JRb, tmp1, sizeof(JRb));
   t3d_dm_mult(rotb, JRg, tmp1, 3, 3, 3, 3); memcpy(JRg, tmp1, sizeof(JRg));
 
   t3d_dm_blocfill(JRea, 9, 3, 9, 1, JRa, JRb, JRg);
   t3d_dm_mult(JeaR, JRea, Jeaea1, 3, 9, 9, 3);
  
   /* translation t = R2.t1 + t2, so dt/dea1 = 0 and dt/dea2 = dR2/dea2 * t1 */
   tmp1[0] = a->matrix.matrix[3];
   tmp1[1] = a->matrix.matrix[7];
   tmp1[2] = a->matrix.matrix[11];

   t3dDerivativeOfEuler(eulerb, JRa, JRb, JRg);
   t3d_dm_mult(JRa, tmp1, Jta, 3, 3, 3, 1);
   t3d_dm_mult(JRb, tmp1, Jtb, 3, 3, 3, 1);
   t3d_dm_mult(JRg, tmp1, Jtg, 3, 3, 3, 1);
   t3d_dm_blocfill(Jtea2, 3, 3, 3, 1, Jta, Jtb, Jtg);

   t3d_dm_mult(JRa, rota, tmp1, 3, 3, 3, 3); memcpy(JRa, tmp1, sizeof(JRa));
   t3d_dm_mult(JRb, rota, tmp1, 3, 3, 3, 3); memcpy(JRb, tmp1, sizeof(JRb));
   t3d_dm_mult(JRg, rota, tmp1, 3, 3, 3, 3); memcpy(JRg, tmp1, sizeof(JRg));
  
   t3d_dm_blocfill(JRea, 9, 3, 9, 1, JRa, JRb, JRg);
   t3d_dm_mult(JeaR, JRea, Jeaea2, 3, 9, 9, 3);
  
   /* the two jacobian matrices */
   t3d_dm_blocfill(Jee1, 6, 6, 3, 3, Jeaea1, NULL, NULL, rotb);
   t3d_dm_blocfill(Jee2, 6, 6, 3, 3, Jeaea2, NULL, Jtea2, t3dmid3); 
   
   /* variance: if C = B o A , V(C) = J^C_A V(A) J^C_A + J^C_B V(B) J^C_B */
   t3d_dmfsf_jmult_6x6(Jee1, a->matrix.v, tmp1); 
   t3d_dmfsf_jmult_6x6(Jee2, b->matrix.v, tmp2); 
   t3d_dms_plus_6x6(tmp1, tmp2, a->matrix.v);

   memcpy(a->matrix.matrix, c, 16*sizeof(double)); 
}
