/*	$LAAS: $ */

/* 
 * Copyright (c) 2000-2004 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#ifdef VXWORKS
#include <vxWorks.h>
#endif

#if defined(__RTAI__) && defined(__KERNEL__)
# include <linux/kernel.h>
# include <linux/sched.h>
#else
# include <stdio.h>
# include <string.h>
#endif

#include <math.h>

#include "libt3d.h"
#include "t3d.h"


/* Local functions --------------------------------------------------- */

static void	convertEulerToVector(T3D *t3d);
static void	convertMatrixToVector(T3D *t3d);


/*
 * --- t3dConvertToVector -----------------------------------------------
 *
 * Convert T3D to vector representation.
 */

void
t3dConvertToVector(T3D *t3d)
{
   switch(t3d->type) {
      case T3D_VECTOR:					break; 
      case T3D_EULER:  convertEulerToVector(t3d);	break;
      case T3D_MATRIX: convertMatrixToVector(t3d);	break;

      default:
	 t3d->any.err = T3D_ENOTIMPLEMENTED; break;
   }
}


/*
 * --- convertEulerToVector ---------------------------------------------
 *
 * Convert Euler T3D to vector representation. Use the simple approach
 * that uses matrix as an intermediate representation. Since the variance
 * matrix is the same for Euler and matrix, this is not that costly.
 */

static void
convertEulerToVector(T3D *t3d)
{
   t3dConvertToMatrix(t3d);
   convertMatrixToVector(t3d);
}


/*
 * --- convertMatrixToVector --------------------------------------------
 *
 * Convert matrix T3D to vector representation.
 */

static void
convertMatrixToVector(T3D *t3d)
{
   double *m = t3d->matrix.matrix;
   double euler[3], cov[21];

   double JvR[27], JRea[36], Jvea[9];
   double JRa[9], JRb[9], JRg[9];

   double rx, ry, rz;
   double ryz, rzx, rxy;
   double dtst, co, t, f, g, max;

#define EPS	(.1)

   /*
    * Transformation
    */

   co = (m[0] + m[5] + m[10] - 1.)/2.0;
   if (co > 1.0) co = 1.0; else if (co < -1.0) co = -1.0;
   t = acos(co);
  
   if ((1+co) > EPS ) {
      /* theta is not close to PI */

      if (t > 1e-10) {
	 f = sin(t)/t;
	 rx = (m[9] - m[6])/2.0/f;
	 ry = (m[2] - m[8])/2.0/f;
	 rz = (m[4] - m[1])/2.0/f;
      } else {
	 /* teta = 0, identity transform */
	 f = 1.0;
	 rx = ry = rz = 0.0;
      }
   } else {
      /* theta is close to PI */

      if (1+co < 1e-10)  {  /* theta == PI */
	 g = (1-co)/t/t; /* g = 2/PI/PI */
	 rx = sqrt((m[0] - co)/g);
	 ry = sqrt((m[5] - co)/g);
	 rz = sqrt((m[10] - co)/g);

	 ryz = (m[9] + m[6])/2.0/g;
	 rzx = (m[2] + m[8])/2.0/g;
	 rxy = (m[4] + m[1])/2.0/g;
      
	 if (rx > ry) max = rx; else max = ry;
	 if (max < rz) max = rz;
      
	 if (rx==max) {
	    if ( rxy < 0.0 ) ry = -ry;
	    if ( rzx < 0.0 ) rz = -rz;
	 } else if (ry==max) {
	    if ( rxy < 0.0 ) rx = -rx;
	    if ( ryz < 0.0 ) rz = -rz;
	 } else {
	    if ( rzx < 0.0 ) rx = -rx;
	    if ( ryz < 0.0 ) ry = -ry;
	 }

      } else {	/* theta close to PI, but not PI */
	 rx = (m[9] - m[6])/2.0;
	 ry = (m[2] - m[8])/2.0;
	 rz = (m[4] - m[1])/2.0;
      
	 if (fabs(rx) > fabs(ry)) max = fabs(rx); else max = fabs(ry);
	 if (max < fabs(rz)) max = fabs(rz);
 
	 if (fabs(rx) == max) {
	    if ( rx < 0.0 ) t = 2*M_PI - t;
	 } else if (fabs(ry) == max) {
	    if ( ry < 0.0 ) t = 2*M_PI - t;
	 } else if (fabs(rz) == max) {
	    if ( rz < 0.0 ) t = 2*M_PI - t;
	 }
      
	 f = sin(t)/t;
	 rx /= f;
	 ry /= f;
	 rz /= f;
      }
   }

   /*
    * Variances
    */

   /* Drives termes vecteur de rotation / termes de la matrice */
   if (fabs(t)< 1e-10) {
      /* theta almost 0, use the limit */
      f=1.0;
      dtst=-1.0/6.0;
   } else {
      if (1+co < 1e-10) {
	 /* theta almost PI */
	 f=1.0;
	 dtst=1.0;
      } else {
	 double dt;

	 f = t/sin(t);
	 dt= -0.5/(sqrt(1.0-(co*co)));
	 dtst= dt*(sin(t)-t*cos(t))/(sin(t)*sin(t));
      }
   }

   t3d_dm_fill(
      JvR, 3, 9,
      dtst*(m[9] - m[6])/2.0, 0.0,                    0.0,
      0.0,                    dtst*(m[9] - m[6])/2.0, -0.5*f,
      0.0,                    0.5*f,                  dtst*(m[9] - m[6])/2.0,

      dtst*(m[2] - m[8])/2.0, 0.0,                    0.5*f,
      0.0,                    dtst*(m[2] - m[8])/2.0, 0.0,
      -0.5*f,                 0.0,                    dtst*(m[2] - m[8])/2.0,

      dtst*(m[4] - m[1])/2.0, -0.5*f,                 0.0,
      0.5*f,                  dtst*(m[4] - m[1])/2.0, 0.0,
      0.0,                    0.0,                    dtst*(m[4] - m[1])/2.0
      );


   /* Drives de la matrice de rotation par raport aux angles d'Euler */
   t3dEulerAnglesOfMatrix(m, euler);
   t3dDerivativeOfEuler(euler, JRa, JRb, JRg);
   t3d_dm_blocfill(JRea, 9, 3, 9, 1, JRa, JRb, JRg);

   /* The resulting jacobian matrix */
   t3d_dm_mult(JvR, JRea, Jvea, 3, 9, 9, 3);
   t3d_dm_blocfill(JRea, 6, 6, 3, 3, Jvea, NULL, NULL, t3dmid3);
   t3d_dmfsf_jmult_6x6(JRea, t3d->matrix.v, cov);

   /* update t3d */
   t3d_dm_fill(t3d->vector.vector, 1, 6, rx, ry, rz, m[3], m[7], m[11]);
   memcpy(t3d->vector.v, cov, sizeof(cov));
   t3d->type = T3D_VECTOR;
}
